@@ -51,9 +51,6 @@
 #include "compiler.h"
 #include "mipointer.h"
 
-/* All drivers implementing backing store need this */
-#include "mibstore.h"
-
 #include "micmap.h"
 
 #include "xf86DDC.h"
@@ -84,6 +81,7 @@
 #endif
 
 #include "omap_driver.h"
+#include "omap_drm.h"
 
 #include "xf86Crtc.h"
 
@@ -111,11 +109,16 @@ typedef struct {
 	struct udev_monitor *uevent_monitor;
 	InputHandlerProc uevent_handler;
 	drmmode_cursor_ptr cursor;
+	int rotated_crtcs;
 } drmmode_rec, *drmmode_ptr;
 
 typedef struct {
 	drmmode_ptr drmmode;
 	drmModeCrtcPtr mode_crtc;
+	Rotation rotation;
+
+	/* properties that we care about: */
+	uint32_t prop_rotation;
 } drmmode_crtc_private_rec, *drmmode_crtc_private_ptr;
 
 typedef struct {
@@ -215,33 +218,50 @@ drmmode_crtc_dpms(xf86CrtcPtr drmmode_crtc, int mode)
 	// FIXME - Implement this function
 }
 
+#define SUPPORTED_ROTATIONS (RR_Rotate_0 | RR_Rotate_90 | RR_Rotate_180 | RR_Rotate_270 | RR_Reflect_X | RR_Reflect_Y)
+
 static Bool
-drmmode_set_mode_major(xf86CrtcPtr crtc, DisplayModePtr mode,
-		Rotation rotation, int x, int y)
+drmmode_set_rotation(xf86CrtcPtr crtc, Rotation rotation)
+{
+#if XF86_CRTC_VERSION >= 4
+	ScrnInfoPtr pScrn = crtc->scrn;
+	drmmode_crtc_private_ptr drmmode_crtc = crtc->driver_private;
+
+	if (!(rotation & ~SUPPORTED_ROTATIONS)) {
+		int ret;
+
+		ret = drmModeObjectSetProperty(drmmode_crtc->drmmode->fd,
+				drmmode_crtc->mode_crtc->crtc_id,
+				DRM_MODE_OBJECT_CRTC,
+				drmmode_crtc->prop_rotation,
+				rotation);
+		if (ret) {
+			ERROR_MSG("failed to set orientation %s", strerror(errno));
+			return FALSE;
+		}
+
+		crtc->driverIsPerformingTransform = TRUE;
+	}
+#endif
+	return xf86CrtcRotate(crtc);
+}
+
+static Bool
+drmmode_restore_crtc(xf86CrtcPtr crtc)
 {
 	ScrnInfoPtr pScrn = crtc->scrn;
 	OMAPPtr pOMAP = OMAPPTR(pScrn);
 	xf86CrtcConfigPtr   xf86_config = XF86_CRTC_CONFIG_PTR(crtc->scrn);
 	drmmode_crtc_private_ptr drmmode_crtc = crtc->driver_private;
 	drmmode_ptr drmmode = drmmode_crtc->drmmode;
-	int saved_x, saved_y;
-	Rotation saved_rotation;
-	DisplayModeRec saved_mode;
 	uint32_t *output_ids = NULL;
 	int output_count = 0;
 	int ret = TRUE;
 	int i;
-	int fb_id;
 	drmModeModeInfo kmode;
 
-	TRACE_ENTER();
-
-	/* remove old fb if it exists */
-	drmmode_remove_fb(pScrn);
-
 	if (drmmode->fb_id == 0) {
-		unsigned int pitch =
-				OMAPCalculateStride(pScrn->virtualX, pScrn->bitsPerPixel);
+		unsigned int pitch = pScrn->displayWidth * (pScrn->bitsPerPixel / 8);
 
 		DEBUG_MSG("create framebuffer: %dx%d",
 				pScrn->virtualX, pScrn->virtualY);
@@ -252,29 +272,15 @@ drmmode_set_mode_major(xf86CrtcPtr crtc, DisplayModePtr mode,
 				pitch, omap_bo_handle(pOMAP->scanout),
 				&drmmode->fb_id);
 		if (ret < 0) {
-			// Fixme - improve this error message:
-			ErrorF("failed to add fb\n");
+			ERROR_MSG("failed to add fb: %s", strerror(errno));
 			return FALSE;
 		}
 	}
 
-	/* Save the current mode in case there's a problem: */
-	saved_mode = crtc->mode;
-	saved_x = crtc->x;
-	saved_y = crtc->y;
-	saved_rotation = crtc->rotation;
-
-	/* Set the new mode: */
-	crtc->mode = *mode;
-	crtc->x = x;
-	crtc->y = y;
-	crtc->rotation = rotation;
-
 	output_ids = calloc(sizeof(uint32_t), xf86_config->num_output);
 	if (!output_ids) {
-		// Fixme - have an error message?
-		ret = FALSE;
-		goto done;
+		ERROR_MSG("allocation failed");
+		return FALSE;
 	}
 
 	for (i = 0; i < xf86_config->num_output; i++) {
@@ -290,29 +296,18 @@ drmmode_set_mode_major(xf86CrtcPtr crtc, DisplayModePtr mode,
 		output_count++;
 	}
 
-	if (!xf86CrtcRotate(crtc))
-		goto done;
-
-	// Fixme - Intel puts this function here, and Nouveau puts it at the end
-	// of this function -> determine what's best for TI'S OMAP4:
-	crtc->funcs->gamma_set(crtc, crtc->gamma_red, crtc->gamma_green,
-			crtc->gamma_blue, crtc->gamma_size);
-
-	drmmode_ConvertToKMode(crtc->scrn, &kmode, mode);
-
-	fb_id = drmmode->fb_id;
+	drmmode_ConvertToKMode(crtc->scrn, &kmode, &crtc->mode);
 
 	ret = drmModeSetCrtc(drmmode->fd, drmmode_crtc->mode_crtc->crtc_id,
-			fb_id, x, y, output_ids, output_count, &kmode);
+			drmmode->fb_id, crtc->x, crtc->y,
+			output_ids, output_count, &kmode);
 	if (ret) {
-		xf86DrvMsg(crtc->scrn->scrnIndex, X_ERROR,
-				"failed to set mode: %s\n", strerror(-ret));
+		ERROR_MSG("failed to set mode: %s", strerror(-ret));
+		ret = FALSE;
 	} else {
 		ret = TRUE;
 	}
 
-	// FIXME - DO WE NEED TO CALL TO THE PVR EXA/DRI2 CODE TO UPDATE THEM???
-
 	/* Turn on any outputs on this crtc that may have been disabled: */
 	for (i = 0; i < xf86_config->num_output; i++) {
 		xf86OutputPtr output = xf86_config->output[i];
@@ -323,17 +318,79 @@ drmmode_set_mode_major(xf86CrtcPtr crtc, DisplayModePtr mode,
 		drmmode_output_dpms(output, DPMSModeOn);
 	}
 
-	// TODO: only call this if we are not using sw cursor.. ie. bad to call this
-	// if we haven't called xf86InitCursor()!!
-	//	if (pScrn->pScreen)
-	//		xf86_reload_cursors(pScrn->pScreen);
-
-done:
 	if (output_ids) {
 		free(output_ids);
 	}
+
+	return ret;
+}
+
+static Bool
+drmmode_set_mode_major(xf86CrtcPtr crtc, DisplayModePtr mode,
+		Rotation rotation, int x, int y)
+{
+	ScrnInfoPtr pScrn = crtc->scrn;
+	drmmode_crtc_private_ptr drmmode_crtc = crtc->driver_private;
+	drmmode_ptr drmmode = drmmode_crtc->drmmode;
+	int saved_x, saved_y;
+	Rotation saved_rotation;
+	DisplayModeRec saved_mode;
+	int ret = TRUE;
+	Bool was_rotated = drmmode->rotated_crtcs > 0;
+
+	TRACE_ENTER();
+
+	/* update the count of number of rotated CRTCs.. if we have one or more
+	 * rotated outputs then we want to use a tiled buffer, but otherwise
+	 * stick with non-tiled
+	 */
+	if ((drmmode_crtc->rotation != RR_Rotate_0) &&
+			(rotation == RR_Rotate_0)) {
+		DEBUG_MSG("disabling rotation for crtc: %u",
+				drmmode_crtc->mode_crtc->crtc_id);
+		drmmode->rotated_crtcs--;
+	} else if ((drmmode_crtc->rotation == RR_Rotate_0) &&
+			(rotation != RR_Rotate_0)) {
+		DEBUG_MSG("enabling rotation for crtc: %u",
+				drmmode_crtc->mode_crtc->crtc_id);
+		drmmode->rotated_crtcs++;
+	}
+
+	drmmode_crtc->rotation = rotation;
+
+	/* at this point, if we are switching from unrotated to rotated
+	 * or visa versa, then we need to reallocate the scanout buffer..
+	 */
+	if (was_rotated != (drmmode->rotated_crtcs > 0)) {
+		/* reallocate scanout buffer.. */
+		drmmode_reallocate_scanout(pScrn, TRUE, crtc);
+	}
+
+	/* note: this needs to be done before setting the mode, otherwise
+	 * drm core will reject connecting the fb to crtc due to mismatched
+	 * dimensions:
+	 */
+	if (!drmmode_set_rotation(crtc, rotation)) {
+		ERROR_MSG("could not set rotation");
+		return FALSE;
+	}
+
+	/* Save the current mode in case there's a problem: */
+	saved_mode = crtc->mode;
+	saved_x = crtc->x;
+	saved_y = crtc->y;
+	saved_rotation = crtc->rotation;
+
+	/* Set the new mode: */
+	crtc->mode = *mode;
+	crtc->x = x;
+	crtc->y = y;
+	crtc->rotation = rotation;
+
+	ret = drmmode_restore_crtc(crtc);
+
 	if (!ret) {
-		/* If there was a problem, resture the old mode: */
+		/* If there was a problem, restore the old mode: */
 		crtc->x = saved_x;
 		crtc->y = saved_y;
 		crtc->rotation = saved_rotation;
@@ -405,6 +462,15 @@ drmmode_show_cursor(xf86CrtcPtr crtc)
 		h = crtc->mode.VDisplay - crtc_y;
 	}
 
+#if XF86_CRTC_VERSION >= 4
+	/* NOTE: driver is taking care of rotation in hw, which means
+	 * we need to deal w/ transformation of mouse cursor ourself:
+	 */
+	if (crtc->driverIsPerformingTransform) {
+		xf86CrtcTransformCursorPos(crtc, &crtc_x, &crtc_y);
+	}
+#endif
+
 	/* note src coords (last 4 args) are in Q16 format */
 	drmModeSetPlane(drmmode->fd, cursor->ovr->plane_id,
 			drmmode_crtc->mode_crtc->crtc_id, cursor->fb_id, 0,
@@ -550,6 +616,7 @@ static const xf86CrtcFuncsRec drmmode_crtc_funcs = {
 static void
 drmmode_crtc_init(ScrnInfoPtr pScrn, drmmode_ptr drmmode, int num)
 {
+	drmModeObjectPropertiesPtr props;
 	xf86CrtcPtr crtc;
 	drmmode_crtc_private_ptr drmmode_crtc;
 
@@ -563,8 +630,23 @@ drmmode_crtc_init(ScrnInfoPtr pScrn, drmmode_ptr drmmode, int num)
 	drmmode_crtc->mode_crtc = drmModeGetCrtc(drmmode->fd,
 			drmmode->mode_res->crtcs[num]);
 	drmmode_crtc->drmmode = drmmode;
+	drmmode_crtc->rotation = RR_Rotate_0;
 
-	// FIXME - potentially add code to allocate a HW cursor here.
+	/* find properties that we care about: */
+	props = drmModeObjectGetProperties(drmmode->fd,
+			drmmode_crtc->mode_crtc->crtc_id, DRM_MODE_OBJECT_CRTC);
+	if (props) {
+		drmModePropertyPtr prop;
+		int i;
+		for (i = 0; i < props->count_props; i++) {
+			prop = drmModeGetProperty(drmmode->fd, props->props[i]);
+			if (!strcmp(prop->name, "rotation")) {
+				drmmode_crtc->prop_rotation = prop->prop_id;
+			}
+			drmModeFreeProperty(prop);
+		}
+		drmModeFreeObjectProperties(props);
+	}
 
 	crtc->driver_private = drmmode_crtc;
 
@@ -1016,55 +1098,123 @@ drmmode_output_init(ScrnInfoPtr pScrn, drmmode_ptr drmmode, int num)
 	return;
 }
 
-static Bool
-drmmode_xf86crtc_resize(ScrnInfoPtr pScrn, int width, int height)
+Bool
+drmmode_is_rotated(ScrnInfoPtr pScrn)
+{
+	OMAPPtr pOMAP = OMAPPTR(pScrn);
+	drmmode_ptr drmmode = drmmode_from_scrn(pScrn);
+	return has_rotation(pOMAP) && drmmode->rotated_crtcs > 0;
+}
+
+Bool
+drmmode_reallocate_scanout(ScrnInfoPtr pScrn, Bool redraw, xf86CrtcPtr crtc)
 {
 	OMAPPtr pOMAP = OMAPPTR(pScrn);
 	ScreenPtr pScreen = pScrn->pScreen;
+	int width = pScrn->virtualX;
+	int height = pScrn->virtualY;
+	Bool rotate = drmmode_is_rotated(pScrn);
 	unsigned int pitch;
 
-	TRACE_ENTER();
+	if (rotate) {
+		pitch = OMAPCalculateTiledStride(width, pScrn->bitsPerPixel);
+	} else {
+		pitch = OMAPCalculateStride(width, pScrn->bitsPerPixel);
+	}
 
-	/* if fb required size has changed, realloc! */
+	if ((width != pOMAP->scanout_w) ||
+			(height != pOMAP->scanout_h) ||
+			(rotate != pOMAP->scanout_rotate)) {
+		uint32_t flags = OMAP_BO_SCANOUT | OMAP_BO_WC;
 
-	DEBUG_MSG("Resize!  %dx%d", width, height);
-
-	pScrn->virtualX = width;
-	pScrn->virtualY = height;
-
-	pitch = OMAPCalculateStride(width, pScrn->bitsPerPixel);
-
-	if ((pitch * height) != omap_bo_size(pOMAP->scanout)) {
-		/* hmm, should we remove fb here.. we don't want to keep
-		 * scanning out a deallocated buffer..
-		 */
+		/* remove old fb if it exists */
 		drmmode_remove_fb(pScrn);
 
+		if (pScreen && pScrn->EnableDisableFBAccess && redraw)
+			pScrn->EnableDisableFBAccess(ENABLE_DISABLE_FB_ACCESS_ARGS(pScrn, FALSE));
+
 		/* delete old scanout buffer */
 		omap_bo_del(pOMAP->scanout);
 
-		DEBUG_MSG("allocating new scanout buffer: %dx%d (%d)",
-				width, height, pitch);
+		if (rotate) {
+			DEBUG_MSG("allocating tiled scanout buffer: %dx%d (%d)",
+					width, height, pitch);
+			flags |= OMAPTiledFlags(pScrn->bitsPerPixel);
+			pOMAP->scanout = omap_bo_new_tiled(pOMAP->dev,
+					width, height, flags);
+		} else {
+			DEBUG_MSG("allocating linear scanout buffer: %dx%d (%d)",
+					width, height, pitch);
+			pOMAP->scanout = omap_bo_new(pOMAP->dev, height * pitch, flags);
+		}
 
-		/* allocate new scanout buffer */
-		pOMAP->scanout = omap_bo_new(pOMAP->dev, height * pitch,
-				OMAP_BO_SCANOUT | OMAP_BO_WC);
 		if (!pOMAP->scanout) {
 			xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 					"Error reallocating scanout buffer\n");
 			return FALSE;
 		}
+
+		pOMAP->scanout_w = width;
+		pOMAP->scanout_h = height;
+		pOMAP->scanout_rotate = rotate;
+
+		pScrn->displayWidth = pitch / (pScrn->bitsPerPixel / 8);
+
+		if (pScreen && pScreen->ModifyPixmapHeader) {
+			PixmapPtr rootPixmap = pScreen->GetScreenPixmap(pScreen);
+			pScreen->ModifyPixmapHeader(rootPixmap,
+					pScrn->virtualX, pScrn->virtualY,
+					pScrn->depth, pScrn->bitsPerPixel, pitch,
+					omap_bo_map(pOMAP->scanout));
+		}
+
+		if (pScreen && pScrn->EnableDisableFBAccess && redraw)
+			pScrn->EnableDisableFBAccess(ENABLE_DISABLE_FB_ACCESS_ARGS(pScrn, TRUE));
+
+		/* if reallocation triggered by a mode-set, we need to reconfigure any
+		 * other crtc's which have just been torn down by destroying the old fb:
+		 */
+		if (crtc) {
+			xf86CrtcConfigPtr config = XF86_CRTC_CONFIG_PTR(pScrn);
+			int i;
+			DEBUG_MSG("restoring CRTCs");
+			for (i = 0; i < config->num_crtc; i++) {
+				if (config->crtc[i] != crtc) {
+					int ret;
+					DEBUG_MSG("restore CRTC %d", i);
+					ret = drmmode_restore_crtc(config->crtc[i]);
+					if (!ret) {
+						ERROR_MSG("failed to reconfig crtc %d", i);
+						/* hmm, I guess just keep trying the rest? */
+					}
+				}
+			}
+		}
+	} else {
+
+		if (pScreen && pScreen->ModifyPixmapHeader) {
+			PixmapPtr rootPixmap = pScreen->GetScreenPixmap(pScreen);
+			pScreen->ModifyPixmapHeader(rootPixmap,
+					pScrn->virtualX, pScrn->virtualY,
+					pScrn->depth, pScrn->bitsPerPixel, pitch,
+					omap_bo_map(pOMAP->scanout));
+		}
 	}
 
-	if (pScreen && pScreen->ModifyPixmapHeader) {
-		PixmapPtr rootPixmap = pScreen->GetScreenPixmap(pScreen);
-		pScreen->ModifyPixmapHeader(rootPixmap,
-				pScrn->virtualX, pScrn->virtualY,
-				pScrn->depth, pScrn->bitsPerPixel, pitch,
-				omap_bo_map(pOMAP->scanout));
-	}
+	return TRUE;
+}
+
+static Bool
+drmmode_xf86crtc_resize(ScrnInfoPtr pScrn, int width, int height)
+{
+	DEBUG_MSG("Resize!  %dx%d", width, height);
+
+	pScrn->virtualX = width;
+	pScrn->virtualY = height;
+
+	if (!drmmode_reallocate_scanout(pScrn, FALSE, NULL))
+		return FALSE;
 
-	TRACE_EXIT();
 	return TRUE;
 }
 
@@ -1338,6 +1488,8 @@ drmmode_wakeup_handler(pointer data, int err, pointer p)
 	if (pScrn == NULL || err < 0)
 		return;
 
+	drmmode = drmmode_from_scrn(pScrn);
+
 	if (FD_ISSET(drmmode->fd, read_mask))
 		drmHandleEvent(drmmode->fd, &event_context);
 }
